/*
  LCD2USB control daemon

  T. Nakagawa
  2011/04/15
*/


#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <syslog.h>
#include <unistd.h>
#include "clcd.h"


#define IDENT "clcdd"
#define DEVICE_SCR "/dev/clcds"
#define DEVICE_KEY "/dev/clcdk"
#define FREQ 30	/* Frequency of screen update/key sense per second */
#define LCD_CONTRAST 200
#define LCD_BRIGHTNESS 255
#define KEYRPT_WAIT 500	/* Time to start key repeat [msec] */
#define KEYRPT_NEXT 12	/* Frequency of key repeat per second */


static int Dirty = 0;
static char Tvram[CLCD_WIDTH * CLCD_HEIGHT];
static int Fd;


void sighandler(int sig) {
  if (sig == SIGTERM) {
    clcd_close();

    syslog(LOG_INFO, "Terminated");

    exit(0);
  } else if (sig == SIGALRM) {
    static char prvkey = '\0';
    static char count;
    char key;
    unsigned char c;

    if (Dirty) {
      clcd_print(Tvram);
      Dirty = 0;
    }
    c = clcd_getkey();
    if (c & 0x01) {
      key = '\x1b';
    } else if (c & 0x02) {
      key = ' ';
    } else if (c & 0x04) {
      key = '\b';
    } else if (c & 0x08) {
      key = '\n';
    } else {
      key = '\0';
    }
    if (key != '\0') {
      if (key == prvkey) {
	if (--count == 0) {
	  write(Fd, &key, sizeof(char));
	  count = FREQ / KEYRPT_NEXT;
	}
      } else {
	write(Fd, &key, sizeof(char));
	count = KEYRPT_WAIT * FREQ / 1000;
      }
    }
    prvkey = key;
  }

  return;
}


int main(void) {
  FILE *fp;
  struct sigaction sa;
  struct itimerval it;

  openlog(IDENT, (LOG_CONS | LOG_PERROR | LOG_PID), LOG_DAEMON);

  mkfifo(DEVICE_SCR, 0666);
  mkfifo(DEVICE_KEY, 0666);

  if ((Fd = open(DEVICE_KEY, O_RDWR | O_NONBLOCK)) < 0) {
    syslog(LOG_ERR, "cannot open %s", DEVICE_KEY);
    return 1;
  }

  if (clcd_open()) {
    syslog(LOG_ERR, "clcd_open() failed: %s", clcd_error);
    return 2;
  }
  clcd_setcontrast(LCD_CONTRAST);
  clcd_setbrightness(LCD_BRIGHTNESS);

  if (daemon(0, 0)) {
    syslog(LOG_ERR, "daemon() failed");
    return 3;
  }

  sa.sa_handler = sighandler;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = SA_RESTART;
  if (sigaction(SIGTERM, &sa, NULL) < 0 || sigaction(SIGALRM, &sa, NULL) < 0) {
    syslog(LOG_ERR, "sigaction() failed");
    return 4;
  }

  it.it_interval.tv_sec = it.it_value.tv_sec = 0;
  it.it_interval.tv_usec = it.it_value.tv_usec = 1000000 / FREQ;
  if (setitimer(ITIMER_REAL, &it, NULL) < 0) {
    syslog(LOG_ERR, "setitimer() failed");
    return 5;
  }

  syslog(LOG_INFO, "Started");

  fp = NULL;
  for (; ; ) {
    char buf[CLCD_WIDTH * CLCD_HEIGHT + 1 + 1];

    if (fp == NULL) {
      fp = fopen(DEVICE_SCR, "r");
      if (fp == NULL) {
	syslog(LOG_ERR, "cannot open %s", DEVICE_SCR);
	return 6;
      }
      syslog(LOG_INFO, "Opened");
    }

    if (fgets(buf, sizeof(buf), fp) == NULL) {
      fclose(fp);
      fp = NULL;
      syslog(LOG_INFO, "Closed");
    } else {
      memcpy(Tvram, buf, sizeof(char) * CLCD_WIDTH * CLCD_HEIGHT);
    }
    Dirty = 1;
  }

  return 0;
}
