/*
  Graphic LCD LQ020B8UB02 control modlue

  T. Nakagawa
  2011/04/09
*/


#include "glcd.h"
#include <ftdi.h>
#include <string.h>
#include <unistd.h>
#include "misaki.xbm"
#include "msk_4x8.xbm"


#define BAUDRATE 1000000
#define BIT_CK 0x01
#define BIT_RES 0x08
#define BIT_A0 0x20
#define BIT_WR 0x40
#define CMD_COLUMN_ADDRESS 0x15
#define CMD_PAGE_ADDRESS 0x75
#define CMD_PIXEL_FORMAT 0xc8
#define CMD_WRITE_RAM 0x5c
#define CMD_DISPLAY_NORMAL 0xa6
#define CMD_SLEEP_DISABLE 0x94
#define CMD_SLEEP_ENABLE 0x95
#define CMD_DISPLAY_OFF 0xae
#define CMD_DISPLAY_ON 0xaf
#define CMD_TIMING_CONTROL 0xca
#define DEV_NOP() {*Buf++ = (BIT_WR | BIT_RES);}
#define DEV_RESET() {*Buf++ = BIT_WR;}
#define DEV_SEND() {ftdi_write_data(&Ftdic, Buf_mem, Buf - Buf_mem); Buf = Buf_mem;}
#define DEV_CMD(cmd) {Buf[0] = (Tbl[(cmd) & 0x0f] | BIT_WR | BIT_RES); Buf[1] = (Buf[0] | BIT_CK); Buf[2] = (Tbl[(cmd) >> 4] | BIT_RES); Buf[3] = (Buf[2] | BIT_WR); Buf += 4;}
#define DEV_DATA(data) {Buf[0] = (Tbl[(data) & 0x0f] | BIT_A0 | BIT_WR | BIT_RES); Buf[1] = (Buf[0] | BIT_CK); Buf[2] = (Tbl[(data) >> 4] | BIT_A0 | BIT_RES); Buf[3] = (Buf[2] | BIT_WR); Buf += 4;}


char *glcd_error;
static struct ftdi_context Ftdic;
static unsigned char Buf_mem[4 * (GLCD_WIDTH * GLCD_HEIGHT * 2 + 11)];
static unsigned char *Buf = Buf_mem;
static const unsigned char Tbl[16] = {
  0x00, 0x10, 0x04, 0x14, 0x02, 0x12, 0x06, 0x16,
  0x80, 0x90, 0x84, 0x94, 0x82, 0x92, 0x86, 0x96,
};
static unsigned short Fgcol = 0xffff;
static unsigned short Bgcol = 0x0000;
static unsigned short Vram[GLCD_WIDTH * GLCD_HEIGHT];


static int dev_open(void) {
  if (ftdi_init(&Ftdic) < 0) {
    glcd_error = "ftdi_init() failed";
    return 1;
  }
  if (ftdi_usb_open(&Ftdic, 0x0403, 0x6001) < 0) {
    glcd_error = "unable to open a ftdi device";
    return 1;
  }
  ftdi_set_baudrate(&Ftdic, BAUDRATE);
  ftdi_set_bitmode(&Ftdic, 0xff, BITMODE_BITBANG);

  return 0;
}


static void dev_close(void) {
  ftdi_disable_bitbang(&Ftdic);
  ftdi_usb_close(&Ftdic);
  ftdi_deinit(&Ftdic);

  return;
}


void glcd_reset(void) {
  /* reset (>20us) & wait (1ms) */
  DEV_RESET();
  DEV_SEND();
  usleep(20);
  DEV_NOP();
  DEV_SEND();
  usleep(1000);

  /* Timing Control & wait (>30ms) */
  DEV_CMD(CMD_TIMING_CONTROL);
  DEV_DATA(0x0c);
  DEV_DATA(0x03);
  DEV_DATA(0x00);
  DEV_DATA(0x0d);
  DEV_DATA(0x03);
  DEV_DATA(0x28);
  DEV_DATA(0x00);
  DEV_SEND();
  usleep(30000);

  /* Sleep Disable & wait (>200ms) */
  DEV_CMD(CMD_SLEEP_DISABLE);
  DEV_SEND();
  usleep(200000);

  return;
}


int glcd_open(void) {
  int rv;

  rv = dev_open();
  if (rv) return rv;

  glcd_reset();

  return 0;
}


void glcd_close(void) {
  DEV_CMD(CMD_DISPLAY_OFF);
  DEV_CMD(CMD_SLEEP_ENABLE);
  DEV_SEND();

  dev_close();

  return;
}


void glcd_putbmp(const unsigned short *bmp) {
  int i, j;

  /* Write RAM command */
  DEV_CMD(CMD_WRITE_RAM);
  for (i = GLCD_HEIGHT - 1; i >= 0; i--) {
    for (j = 0; j < GLCD_WIDTH; j++) {
      DEV_DATA(bmp[GLCD_WIDTH * i + j] >> 8);
      DEV_DATA(bmp[GLCD_WIDTH * i + j] & 0xff);
    }
  }

  /* Display Normal command */
  DEV_CMD(CMD_DISPLAY_NORMAL);

  /* Dispaly On command */
  DEV_CMD(CMD_DISPLAY_ON);

  DEV_SEND();

  return;
}


void glcd_fgcolor(unsigned char r, unsigned char g, unsigned char b) {
  Fgcol = (((unsigned short)r << 11) | ((unsigned short)(g & 0x3f) << 5) | (b & 0x1f));

  return;
}


void glcd_bgcolor(unsigned char r, unsigned char g, unsigned char b) {
  r >>= 3;
  g >>= 2;
  b >>= 3;
  Bgcol = (((unsigned short)r << 11) | ((unsigned short)g << 5) | b);

  return;
}


void glcd_print(int x, int y, const char *str) {
  int i;

  while (*str != '\0') {
    if (!(*str & 0x80)) {
      int row, col;
      int ptr;
      int pos;

      row = (unsigned char)*str / 16;
      col = (unsigned char)*str++ % 16;
      ptr = row * msk_4x8_width / 8 * 8 + col / 2;
      pos = GLCD_WIDTH * y + x;
      for (i = 0; i < 8; i++) {
	unsigned char pat;
	pat = msk_4x8_bits[ptr];
	if (col % 2 == 1) pat >>= 4;
	Vram[pos + 0] = (pat & 0x01) ? Fgcol : Bgcol;
	Vram[pos + 1] = (pat & 0x02) ? Fgcol : Bgcol;
	Vram[pos + 2] = (pat & 0x04) ? Fgcol : Bgcol;
	Vram[pos + 3] = (pat & 0x08) ? Fgcol : Bgcol;
	ptr += msk_4x8_width / 8;
	pos += GLCD_WIDTH;
      }
      x += 4;
    } else {
      int row, col;
      int ptr;
      int pos;

      row = (unsigned char)*str++ - 0xa1;
      col = (unsigned char)*str++ - 0xa1;
      ptr = row * misaki_width / 8 * 8 + col;
      pos = GLCD_WIDTH * y + x;
      for (i = 0; i < 8; i++) {
	unsigned char pat;
	pat = misaki_bits[ptr];
	Vram[pos + 0] = (pat & 0x01) ? Fgcol : Bgcol;
	Vram[pos + 1] = (pat & 0x02) ? Fgcol : Bgcol;
	Vram[pos + 2] = (pat & 0x04) ? Fgcol : Bgcol;
	Vram[pos + 3] = (pat & 0x08) ? Fgcol : Bgcol;
	Vram[pos + 4] = (pat & 0x10) ? Fgcol : Bgcol;
	Vram[pos + 5] = (pat & 0x20) ? Fgcol : Bgcol;
	Vram[pos + 6] = (pat & 0x40) ? Fgcol : Bgcol;
	Vram[pos + 7] = (pat & 0x80) ? Fgcol : Bgcol;
	ptr += misaki_width / 8;
	pos += GLCD_WIDTH;
      }
      x += 8;
    }
  }

  return;
}


void glcd_clear(void) {
  int i;

  for (i = 0; i < GLCD_WIDTH * GLCD_HEIGHT; i++) {
    Vram[i] = Bgcol;
  }

  return;
}


void glcd_scroll(int size) {
  int i;

  if (size > 0) {
    memmove(Vram, Vram + size, sizeof(unsigned short) * (GLCD_WIDTH * GLCD_HEIGHT - size));
    for (i = GLCD_WIDTH * GLCD_HEIGHT - size; i < GLCD_WIDTH * GLCD_HEIGHT; i++) {
      Vram[i] = Bgcol;
    }
  } else {
    size = -size;
    memmove(Vram + size, Vram, sizeof(unsigned short) * (GLCD_WIDTH * GLCD_HEIGHT - size));
    for (i = 0; i < size; i++) {
      Vram[i] = Bgcol;
    }
  }

  return;
}


void glcd_refresh(void) {
  int i;

  /* Write RAM command */
  DEV_CMD(CMD_WRITE_RAM);
  for (i = 0; i < GLCD_WIDTH * GLCD_HEIGHT; i++) {
    DEV_DATA(Vram[i] >> 8);
    DEV_DATA(Vram[i] & 0xff);
  }

  /* Display Normal command */
  DEV_CMD(CMD_DISPLAY_NORMAL);

  /* Dispaly On command */
  DEV_CMD(CMD_DISPLAY_ON);

  DEV_SEND();

  return;
}


void glcd_font(char *chr, char *font) {
  int i;
  int row, col;
  int ptr;

  row = (unsigned char)*chr++ - 0xa1;
  col = (unsigned char)*chr++ - 0xa1;
  ptr = row * misaki_width / 8 * 8 + col;
  for (i = 0; i < 8; i++) {
    misaki_bits[ptr] = font[i];
    ptr += misaki_width / 8;
  }

  return;
}
