/*
  Graphic LCD LQ020B8UB02 control daemon

  T. Nakagawa
  2011/04/13
*/


#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <syslog.h>
#include <unistd.h>
#include "glcd.h"


#define IDENT "glcdd"
#define DEVICE "/dev/glcds"
#define FREQ 30	/* Frequency of screen update per second */
#define TXT_WIDTH 32
#define TXT_HEIGHT 20


static int Dirty = 0;
static int Mode = 0;
static unsigned short Gvram[GLCD_WIDTH * GLCD_HEIGHT];


void sighandler(int sig) {
  if (sig == SIGTERM) {
    glcd_close();

    syslog(LOG_INFO, "Terminated");

    exit(0);
  } else if (sig == SIGALRM) {
    if (Dirty) {
      if (!Mode) {
	glcd_refresh();
      } else {
	glcd_putbmp(Gvram);
      }
      Dirty = 0;
    }
  }

  return;
}


int main(void) {
  FILE *fp;
  struct sigaction sa;
  struct itimerval it;
  int x, y;

  openlog(IDENT, (LOG_CONS | LOG_PERROR | LOG_PID), LOG_DAEMON);

  mkfifo(DEVICE, 0666);

  if (glcd_open()) {
    syslog(LOG_ERR, "glcd_open() failed: %s", glcd_error);
    return 1;
  }

  if (daemon(0, 0)) {
    syslog(LOG_ERR, "daemon() failed");
    return 2;
  }

  sa.sa_handler = sighandler;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = SA_RESTART;
  if (sigaction(SIGTERM, &sa, NULL) < 0 || sigaction(SIGALRM, &sa, NULL) < 0) {
    syslog(LOG_ERR, "sigaction() failed");
    return 3;
  }

  it.it_interval.tv_sec = it.it_value.tv_sec = 0;
  it.it_interval.tv_usec = it.it_value.tv_usec = 1000000 / FREQ;
  if (setitimer(ITIMER_REAL, &it, NULL) < 0) {
    syslog(LOG_ERR, "setitimer() failed");
    return 4;
  }

  syslog(LOG_INFO, "Started");

  fp = NULL;
  x = 0;
  y = 0;
  for (; ; ) {
    int c;

    if (fp == NULL) {
      fp = fopen(DEVICE, "r");
      if (fp == NULL) {
	syslog(LOG_ERR, "cannot open %s", DEVICE);
	return 5;
      }
      syslog(LOG_INFO, "Opened");
    }

    c = getc(fp);
    if (c == EOF) {
      fclose(fp);
      fp = NULL;
      syslog(LOG_INFO, "Closed");
    } else if (c == '\x01') {
      /* Transfer bitmap */
      fread(Gvram, sizeof(unsigned short), GLCD_WIDTH * GLCD_HEIGHT, fp);
    } else if (c == '\x04') {
      /* Set font */
      char buf[10];
      fread(buf, sizeof(char), 10, fp);
      glcd_font(buf, buf + 2);
    } else if (c == '\x05') {
      /* Set color */
      unsigned char col[6];
      fread(col, sizeof(unsigned char), 6, fp);
      glcd_fgcolor(col[0], col[1], col[2]);
      glcd_bgcolor(col[3], col[4], col[5]);
    } else if (c == '\x06') {
      /* Move cursor */
      unsigned char buf[2];
      fread(buf, sizeof(unsigned char), 2, fp);
      x = buf[0];
      y = buf[1];
    } else if (c == '\x0a') {	/* '\n' */
      /* Line feed */
      x = 0;
      y++;
      if (y >= TXT_HEIGHT) {
	y--;
	glcd_scroll(GLCD_WIDTH * GLCD_FONTHEIGHT);
      }
    } else if (c == '\x0c') {	/* '\f' */
      /* Move to home & clear (necessary for dumb terminal) */
      x = 0;
      y = 0;
      glcd_fgcolor(0, 0, 0);
      glcd_bgcolor(255, 255, 255);
      glcd_clear();
    } else if (c == '\x0d') {	/* '\r' */
      /* Carriage return */
      x = 0;
    } else if (c == '\x0e') {
      /* Switch to graphic mode */
      Mode = 1;
    } else if (c == '\x0f') {
      /* Switch to text mode */
      Mode = 0;
    } else if (c == '\x1b') {
      /* Move cursor (necessary for dumb terminal) "\x1b[<line>;<col>H" */
      int col, lin;
      fscanf(fp, "[%d;%dH", &lin, &col);
      lin--;
      col--;
      if (col >= 0 && col < TXT_WIDTH) x = col;
      if (lin >= 0 && lin < TXT_HEIGHT) y = lin;
    } else {
      /* ASCII/EUC-JP character */
      char buf[3];
      if (!(c & 0x80)) {
	if (x + 1 > TXT_WIDTH) {
	  x = 0;
	  y++;
	  if (y >= TXT_HEIGHT) {
	    y--;
	    glcd_scroll(GLCD_WIDTH * GLCD_FONTHEIGHT);
	  }
	}
	buf[0] = c;
	buf[1] = '\0';
	glcd_print(GLCD_FONTWIDTH * x, GLCD_FONTHEIGHT * y, buf);
	x += 1;
      } else {
	if (x + 2 > TXT_WIDTH) {
	  x = 0;
	  y++;
	  if (y >= TXT_HEIGHT) {
	    y--;
	    glcd_scroll(GLCD_WIDTH * GLCD_FONTHEIGHT);
	  }
	}
	buf[0] = c;
	buf[1] = getc(fp);
	buf[2] = '\0';
	glcd_print(GLCD_FONTWIDTH * x, GLCD_FONTHEIGHT * y, buf);
	x += 2;
      }
    }
    Dirty = 1;
  }

  return 0;
}
