/*
  LCD Shell

  T. Nakagawa
  2011/04/16
*/


#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/wait.h>
#include <unistd.h>


#define WIDTH 32
#define HEIGHT 20
#define DIR_MAX 1024	/* Maximum length of path string */
#define FILE_MAX ((HEIGHT - 2) * 20)
#define EXT_MAX 256	/* Maximum number of extensions */
#define HASH_SIZE 1023	/* Hash table size */
#define HASH_INC 10
#define TYPE_MAX 16	/* Maximum number of file types */
#define CLR_HEADLN "\xff\xff\xff\x00\x00\xff"
#define CLR_CURSOR "\x00\x00\x00\xff\xff\x00"
#define CLR_NORMAL "\x00\x00\x00\xff\xff\xff"
#define CLR_STATUS "\xff\xff\xff\x00\xc0\x00"
#define PATH "/usr/local/lib/lcdshell"


struct {
  unsigned char icon[8];
  char *ext;
  char *prog;
} Ftype[] = {
  {{0x0e, 0x11, 0x7f, 0x41, 0x41, 0x41, 0x7f, 0x00}, ".\0", PATH "/lcdshell"},
  {{0x1f, 0x21, 0x2d, 0x21, 0x2d, 0x21, 0x3f, 0x00}, "txt\0", PATH "/pager.sh"},
  {{0x2a, 0x3e, 0x3e, 0x1c, 0x49, 0x6b, 0x1c, 0x00}, "jpg\0jpeg\0png\0bmp\0", PATH "/imgviewer.sh"},
  {{0x30, 0x70, 0x50, 0x50, 0x5c, 0x1e, 0x0e, 0x00}, "wav\0mp3\0m3u\0pls\0", PATH "/mplayer.sh"},
  {{0x7f, 0x71, 0x67, 0x49, 0x53, 0x55, 0x7f, 0x00}, "rss\0", PATH "/rssreader.sh"},
  {{0x00, 0x3e, 0x22, 0x22, 0x3e, 0x41, 0x7f, 0x00}, "exe\0", PATH "/launcher.sh"},
  {{0x0e, 0x1c, 0x38, 0x38, 0x38, 0x1c, 0x0e, 0x00}, "scr\0", PATH "/ssaver.sh"},
};

struct {
  int num;
  char *name[EXT_MAX];
  int type[EXT_MAX];
  int ht[HASH_SIZE];
} Ext;


unsigned int strhash(const char *str) {
  unsigned int h = 0;

  while (*str != '\0') h = *(unsigned char *)str++ + 31 * h;

  return h;
}


int dir_filter(const struct dirent *dp) {
  unsigned int h;
  char *ext;

  if (dp->d_name[0] == '.') return 0;
  if (dp->d_type == DT_DIR) {
    ext = ".";
  } else {
    ext = rindex(dp->d_name, '.');
    if (ext == NULL) {
      ext = "";
    } else {
      ext++;
    }
  }
  for (h = strhash(ext) % HASH_SIZE; Ext.ht[h] >= 0; h = (h + HASH_INC) % HASH_SIZE) if (strcmp(ext, Ext.name[Ext.ht[h]]) == 0) break;
  if (Ext.ht[h] < 0) return 0;

  return Ext.type[Ext.ht[h]] + 1;
}


int dir_compar(const void *a, const void *b) {
  int a_type, b_type;
  char *a_name, *b_name;

  a_type = dir_filter(*(struct dirent **)a);
  b_type = dir_filter(*(struct dirent **)b);
  if (a_type != b_type) return a_type - b_type;

  a_name = (*(struct dirent **)a)->d_name;
  b_name = (*(struct dirent **)b)->d_name;
  return strcmp(a_name, b_name);

  return 0;
}


void call(int type, const char *dir) {
  pid_t pid;

  write(1, "\x05\x00\x00\x00\xff\xff\xff\f\x0f", 9);	/* Reset LCD for text */

  if ((pid = fork()) < 0) return;
  if (pid != 0) {
    wait(NULL);
  } else {
    execl(Ftype[type].prog, Ftype[type].prog, dir, NULL);
    exit(1);
  }

  return;
}


int main(int argc, char **argv) {
  int i;
  unsigned int h;
  char *dir;
  struct {
    int num;
    char name[FILE_MAX][WIDTH + 1 + 1];
    int type[FILE_MAX];
    struct dirent **list;
  } file;
  int curbgn, curpos;
  int dirty;

  if (argc != 2) return 1;
  dir = argv[1];

  if (dir[0] != '/' || strlen(dir) > DIR_MAX) return 1;

  /* Register file extensions to a hash table */
  for (i = 0; i < HASH_SIZE; i++) Ext.ht[i] = -1;
  Ext.num = 0;
  for (i = 0; i < sizeof(Ftype) / sizeof(*Ftype) && i < TYPE_MAX && Ext.num < EXT_MAX; i++) {
    char *p;
    char buf[3];
    for (p = Ftype[i].ext; *p != '\0'; p += strlen(p) + 1) {
      Ext.name[Ext.num] = p;
      Ext.type[Ext.num] = i;
      for (h = strhash(p) % HASH_SIZE; Ext.ht[h] >= 0; h = (h + HASH_INC) % HASH_SIZE) ;
      Ext.ht[h] = Ext.num;
      Ext.num++;
    }
    buf[0] = 0x04;
    buf[1] = 0xfe;
    buf[2] = 0xa1 + i;
    write(1, buf, 3);
    write(1, Ftype[i].icon, 8);
  }

 restart:

  /* Read file names */
  file.num = scandir(dir, &file.list, dir_filter, dir_compar);
  if (file.num <= 0) return 1;
  if (file.num > FILE_MAX) file.num = FILE_MAX;
  for (i = 0; i < file.num; i++) {
    char *tmp;
    int n;
    file.type[i] = dir_filter(file.list[i]) - 1;
    tmp = rindex(file.list[i]->d_name, '.');
    if (tmp != NULL) *tmp = '\0';
    sprintf(file.name[i], "%c%c%-30.30s\n", 0xfe, 0xa1 + file.type[i], file.list[i]->d_name);
    if (tmp != NULL) *tmp = '.';
    for (tmp = &file.name[i][2], n = 0; *tmp != '\0'; tmp++) if ((*tmp & 0x80) != 0) n++;
    if (n % 2 != 0) file.name[i][31] = ' ';
  }
  for (; i % (HEIGHT - 2) != 0; i++) {
    strcpy(file.name[i], "                                \n");
  }

  curbgn = 0;
  curpos = 0;
  dirty = 1;
  for (; ; ) {
    char c;
    unsigned char n;

    if (dirty) {
      char *dirname;
      char buf[WIDTH + 1 + 1];

      write(1, "\x05\x00\x00\x00\xff\xff\xff\f\x0f", 9);	/* Reset LCD for text */
      /* Title line */
      write(1, "\x05" CLR_HEADLN, 7);
      dirname = rindex(dir, '/') + 1;
      if (strlen(dirname) > WIDTH) dirname = "";
      sprintf(buf, "%-32s\n", dirname);
      write(1, buf, WIDTH);
      /* File list */
      write(1, "\x05" CLR_NORMAL, 7);
      for (i = 0; i < HEIGHT - 2; i++) {
	if (i == curpos - curbgn) {
	  write(1, "\x05" CLR_CURSOR, 7);
	  write(1, file.name[curbgn + i], WIDTH + 1);
	  write(1, "\x05" CLR_NORMAL, 7);
	} else {
	  write(1, file.name[curbgn + i], WIDTH + 1);
	}
      }
      /* Status line */
      write(1, "\x05" CLR_STATUS, 7);
      sprintf(buf, "%3d files                [%2d/%2d]", file.num, curbgn / (HEIGHT - 2) + 1, (file.num - 1) / (HEIGHT - 2) + 1);
      write(1, buf, WIDTH);
      dirty = 0;
    }

    if (read(0, &c, 1) != 1) break;

    if (c == '\x1b') {
      if (argv[0][0] == '-') goto restart;
      break;	/* Exit if not a login shell */
    } else if (c == '\n') {
      char newdir[DIR_MAX + 1];

      if (strlen(dir) + 1 + strlen(file.list[curpos]->d_name) > DIR_MAX) continue;
      sprintf(newdir, "%s/%s", dir, file.list[curpos]->d_name);
      call(file.type[curpos], newdir);
      dirty = 1;
    } else if (c == ' ' && curpos + 1 < file.num) {
      curpos++;
      if (curpos - curbgn >= HEIGHT - 2) {
	curbgn += HEIGHT - 2;
	dirty = 1;
      } else {
	write(1, "\x06\x00", 2);
	n = 1 + (curpos - 1) - curbgn;
	write(1, &n, 1);
	write(1, "\x05" CLR_NORMAL, 7);
	write(1, file.name[curpos - 1], WIDTH + 1);
	write(1, "\x05" CLR_CURSOR, 7);
	write(1, file.name[curpos], WIDTH + 1);
      }
    } else if (c == '\b' && curpos - 1 >= 0) {
      curpos--;
      if (curpos - curbgn < 0) {
	curbgn -= HEIGHT - 2;
	dirty = 1;
      } else {
	write(1, "\x06\x00", 2);
	n = 1 + curpos - curbgn;
	write(1, &n, 1);
	write(1, "\x05" CLR_CURSOR, 7);
	write(1, file.name[curpos], WIDTH + 1);
	write(1, "\x05" CLR_NORMAL, 7);
	write(1, file.name[curpos + 1], WIDTH + 1);
      }
    }
  }

  return 0;
}
